print("import: 'genshi'")
import genshi

print("import: 'genshi.filters'")
import genshi.filters

print("import: 'genshi.filters.tests'")
import genshi.filters.tests

print("import: 'genshi.template'")
import genshi.template

print("import: 'genshi.template.tests'")
import genshi.template.tests

print("import: 'genshi.template.tests.templates'")
import genshi.template.tests.templates

print("import: 'genshi.tests'")
import genshi.tests

